<?php
/**
 * The core plugin class.
 *
 * This is used to define internationalization, admin-specific hooks, and public-facing site hooks.
 * Also maintains the unique identifier of this plugin as well as the current
 * version of the plugin.
 *
 * @since      1.0.0
 * @package    hassle-free-fonts
 * @author     Syncment <info@syncment.nl>
 */
class Hassle_Free_Fonts
{
    //--------------------------------------------------------------------------------------------------------------------------------

    /**
     * Instance of this class.
     *
     * @var      object
     */
    protected static $instance = null;

	/**
     * Return an instance of this class.
     *
     * @return    object    A singleton instance of this class.
     */
    public static function get_instance() {

        // If the single instance hasn't been set, set it now.
        if ( null == self::$instance ) {
            self::$instance = new self;
        }
        return self::$instance;
    }

    //--------------------------------------------------------------------------------------------------------------------------------
    
    /**
     * 
     */
    public function activate() 
    {
        // create upload/download directory
		if (!file_exists($this->output['path'])) {
			mkdir($this->output['path'], 0755, true);
		}
		
		// create upload/download directory
		if (!file_exists($this->output['path'] . '/fonts')) {
			mkdir($this->output['path'] . '/fonts', 0755, true);
        }
        
        // create public css file
		if (!file_exists($this->output['path'] . '/style.css')) {
			$css = fopen($this->output['path'] . '/style.css', "w");
			fwrite($css, "/* Hassle-Free Fonts - Wordpress plugin */");
			fclose($css);
		}
    }

    /**
     * 
     */
    public function deactivate() 
    {

    }

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * 
     */
	private $plugin_name = 'hassle-free-fonts';
	
	/**
     * 
     */
	private $version = '1.0.0';
	
	/**
	 * 
	 */
	private $output = array();

	/**
	 * 
	 */
	private $permalink;

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
	 * 
	 */
	public function __construct() 
	{
        $this->output = $this->define_output_directions($this->plugin_name);
    }
    
    /**
     * 
     */
    public function init() 
    {
        //
        $this->define_admin_hooks();
        $this->define_public_hooks(); 
    }

    // -------------------------------------------------------------------------------------------------------------------------------

    /**
     * 
     */
    private function define_admin_hooks() 
    {
        // create custom 
        add_action('admin_menu', array( $this, 'custom_admin_appearance_page' ));

        // define rest api for react
        add_action( 'rest_api_init', array( $this, 'create_rest_api_hook' ) );
    }

    /**
     * 
     */
    private function define_public_hooks() 
    {
        // hook fontend css file
        add_action('wp_enqueue_scripts', array( $this, 'public_enqueue_styles' ));
    }

    //--------------------------------------------------------------------------------------------------------------------------------
    // Private

    /**
	 * Create custom appearance page
	 */
	function custom_admin_appearance_page()
	{
		// create admin page
		$menu = add_theme_page( 'Fonts', 'Fonts', 'manage_options', $this->plugin_name, array( $this, 'custom_admin_content_callback' ) );

		//
		$this->permalink = get_option( 'permalink_structure' );

		//
		if(isset($this->permalink) && $this->permalink != null)
		{		
			// add styles
			add_action( 'admin_print_styles-' . $menu, array( $this, 'admin_enqueue_styles' ) );

			// add scripts
			add_action( 'admin_print_scripts-' . $menu, array( $this, 'admin_enqueue_scripts' ) );	
		}
	}
	
	/**
	 * Load backend styles
	 */
	function admin_enqueue_styles() 
	{
		// load css admin style
		wp_enqueue_style( $this->plugin_name, plugin_dir_url( dirname( __FILE__ ) ) . 'admin/style.css', array(), $this->version, 'all' );
	}

    /**
	 * Load backend scripts
	 */
	function admin_enqueue_scripts() 
	{
		// load react script
		wp_enqueue_script( $this->plugin_name, plugin_dir_url( dirname( __FILE__ ) ) . 'admin/script.js', array( 'wp-element', 'wp-polyfill' ), $this->version, true );

		// localize data for react
		wp_localize_script( $this->plugin_name, 'SYNC', array(
			'url' 	=> plugin_dir_url( dirname( __FILE__ ) ),
            'rest' 	=> esc_url_raw( rest_url() ),
			'nonce' => wp_create_nonce( 'wp_rest' )
		));
		
		// load wordpress code editor
		$code_editor_settings = wp_enqueue_code_editor( array( 'type' => 'text/css' ) );
	}

	/**
	 * Create page content (react placeholder)
	 */
	function custom_admin_content_callback()
	{
		//
		if($this->permalink == "") {
			//
			echo '<div class="notice notice-error"><p>Permalink is required to use Wordpress REST API, please set an <a href="' . admin_url('options-permalink.php') . '">permalink here</a>.</p></div>';
		}
		else {
			//
			echo '<div id="syncment_react_content"></div>';
		}
	}

    //--------------------------------------------------------------------------------------------------------------------------------
    // Public

    /**
	 * Load public css file
	 */
	function public_enqueue_styles() 
	{
		wp_enqueue_style( $this->plugin_name, $this->output['url'] . '/style.css', array(), $this->version, 'all' );
    }

    //--------------------------------------------------------------------------------------------------------------------------------
    // REST API

    /**
	 * 
	 */
	function create_rest_api_hook()
	{
		// get current user for securing rest api
		$this->current_user = wp_get_current_user();

		//--------------------------------------------------------

		// GET: /wp-json/hassle-free-fonts/v1/files
		register_rest_route( $this->plugin_name . '/v1', '/files/', array(
			'methods'  => 'get',
			'callback' => array( $this, 'rest_get_files_callback'),
			'permission_callback' => function() {
                return user_can($this->current_user, 'manage_options');
            }
		));

		// DELETE: /wp-json/hassle-free-fonts/v1/files
		register_rest_route( $this->plugin_name . '/v1', '/files/', array(
			'methods'  => 'post',
			'callback' => array( $this, 'rest_delete_files_callback'),
			'permission_callback' => function() {
                return user_can($this->current_user, 'manage_options');
            }
		));

		//--------------------------------------------------------

		// GET: /wp-json/hassle-free-fonts/v1/style
		register_rest_route( $this->plugin_name . '/v1', '/style/', array(
			'methods'  => 'get',
			'callback' => array( $this, 'rest_get_style_callback'),
			'permission_callback' => function() {
                return user_can($this->current_user, 'manage_options');
            }
		));

		// GET: /wp-json/hassle-free-fonts/v1/style
		register_rest_route( $this->plugin_name . '/v1', '/style/', array(
			'methods'  => 'put',
			'callback' => array( $this, 'rest_put_style_callback'),
			'permission_callback' => function() {
                return user_can($this->current_user, 'manage_options');
            }
		));

		// GET: /wp-json/hassle-free-fonts/v1/style/generate
		register_rest_route( $this->plugin_name . '/v1', '/style/generate/', array(
			'methods'  => 'get',
			'callback' => array( $this, 'rest_get_style_generate_callback'),
			'permission_callback' => function() {
                return user_can($this->current_user, 'manage_options');
            }
		));

		// POST: /wp-json/hassle-free-fonts/v1/style
		register_rest_route( $this->plugin_name . '/v1', '/style/', array(
			'methods'  => 'post',
			'callback' => array( $this, 'rest_post_style_callback'),
			'permission_callback' => function() {
                return user_can($this->current_user, 'manage_options');
            }
		));

		//--------------------------------------------------------

		// GET: /wp-json/hassle-free-fonts/v1/fonts
		register_rest_route( $this->plugin_name . '/v1', '/fonts/', array(
			'methods'  => 'get',
			'callback' => array( $this, 'rest_get_fonts_callback'),
			'permission_callback' => function() {
                return user_can($this->current_user, 'manage_options');
            }
		));

	    // POST: /wp-json/hassle-free-fonts/v1/fonts
		register_rest_route( $this->plugin_name . '/v1', '/fonts/', array(
			'methods'  => 'post',
			'callback' => array( $this, 'rest_post_fonts_callback'),
			'permission_callback' => function() {
                return user_can($this->current_user, 'manage_options');
            }
		));

		//--------------------------------------------------------
    }
    
    /**
	 * GET: /wp-json/hassle-free-fonts/v1/files
     * 
     * Get current files from output directory
     * 
	 */
	function rest_get_files_callback()
	{	
		//
		$files = $this->get_fonts_files();

		//
        return new WP_REST_Response( array( 'message' => "...", 'files' => $files ), 200 );
	}

	/**
	 * DELETE: /wp-json/hassle-free-fonts/v1/files
     * 
     * Get current files from output directory
     * 
	 */
	function rest_delete_files_callback(WP_REST_Request $request)
	{	
		// double check nonce
		if ( ! wp_verify_nonce( $request->get_header('X-WP-Nonce'), 'wp_rest' ) ) {
			throw new Exception('Wordpress nonce is invalid.');
		}           

		// getting form post style
		$id = $request->get_param('id');
		$weight = $request->get_param('weight');
		$style = $request->get_param('style');

		//
		$files = $this->get_fonts_files();

		//
		$wp_upload = wp_get_upload_dir();
		$files_dir = $wp_upload['basedir'] . '/' . $this->plugin_name . '/fonts/';
		
		//
		foreach($files[$id][$weight][$style] as $file)
		{	
			//
			unlink($files_dir . $file);
		}

		//
		$files = $this->get_fonts_files();

		//
        return new WP_REST_Response( array( 'message' => "Font successfully deleted.", 'files' => $files ), 200 );
	}

    /**
	 * GET: /wp-json/hassle-free-fonts/v1/style
     * 
     * Get content of public css file
     * 
	 */
	function rest_get_style_callback()
	{
		//
		return new WP_REST_Response( array( 'url'  => $this->output['url'] . '/style.css', 'content' => file_get_contents($this->output['path'] . '/style.css') ), 200 );
	}

	/**
	 * PUT: /wp-json/hassle-free-fonts/v1/style
     * 
     * 
     * 
	 */
	function rest_put_style_callback(WP_REST_Request $request)
	{
		try
		{
			// double check nonce
			if ( ! wp_verify_nonce( $request->get_header('X-WP-Nonce'), 'wp_rest' ) ) {
				throw new Exception('Wordpress nonce is invalid.');
            }           

			//
			$style = $this->generate_style();

			// open file to write content
			$css_file = fopen($this->output['path'] . '/style.css', "w");

			// write content			
			fwrite($css_file, $style);

			// close file
			fclose($css_file);

			//
			return new WP_REST_Response( array( 'message' => "CSS generated and saved successfully." ), 200 );	
		}
		catch (Exception $ex) 
		{
			return new WP_REST_Response( array( 'message' => $ex->getMessage(), ), 400 );		
		}
	}

    /**
	 * GET: /wp-json/hassle-free-fonts/v1/style/generate
     * 
     * Generate new CSS file through found font files.
     * 
	 */
	function rest_get_style_generate_callback()
	{
		//
		$style = $this->generate_style();
		
		//
		return new WP_REST_Response( array( 'style' => $style ), 200 );
	}

	/**
	 * 
	 */
	function rest_post_style_callback(WP_REST_Request $request)
	{
		try
		{
			// double check nonce
			if ( ! wp_verify_nonce( $request->get_header('X-WP-Nonce'), 'wp_rest' ) ) {
				throw new Exception('Wordpress nonce is invalid.');
            }           

			// getting form post style
			$style = $request->get_param('style');

			// open file to write content
			$css_file = fopen($this->output['path'] . '/style.css', "w");

			// write content			
			fwrite($css_file, isset($style) && $style != "" ? stripslashes( $style ) : "");

			// close file
			fclose($css_file);

			//
			return new WP_REST_Response( array( 'message' => "Public style saved successfully" ), 200 );	
		}
		catch (Exception $ex) 
		{
			return new WP_REST_Response( array( 'message' => $ex->getMessage(), ), 400 );		
		}
	}

    /**
	 * GET: /wp-json/hassle-free-fonts/v1/fonts
     * 
     * Get fonts from remote API, this is like a proxy.
     * 
	 */
	function rest_get_fonts_callback()
	{
		$response = wp_remote_get("https://google-webfonts-helper.herokuapp.com/api/fonts");
		echo wp_remote_retrieve_body($response);
		exit;
	}

	/**
	 * 
	 */
	function rest_post_fonts_callback(WP_REST_Request $request)
	{
		try
		{
			// double check nonce
			if ( ! wp_verify_nonce( $request->get_header('X-WP-Nonce'), 'wp_rest' ) ) {
				throw new Exception('Wordpress nonce is invalid.');
			}

			// define form data
            $font =     $request->get_param('font');
            $family =   $request->get_param('family');
			$subset =   $request->get_param('subset');
			$variants = $request->get_param('variants');

			//
			if(!isset($font) || $font === '') {
				throw new Exception('Font is required for downloading.');
			}			

			// download zip file
			$zip_file = $this->download_files( $font, $subset, $variants );	

			// unpack zip file
			$unzipped_files = $this->unzip_file($zip_file);

			//
			$data = array(
                'message'   => "Font has been successfully downloaded and extracted.",
				'font'      => $font,
				'family'    => $family,
				'subset'    => $subset,
				'variants' 	=> $variants,
				'files'		=> $unzipped_files
			);

			//
			return new WP_REST_Response( $data, 200 );
		}
		catch (Exception $ex) 
		{
			return new WP_REST_Response( array( 'message' => $ex->getMessage(), ), 400 );		
		}
	}


    //--------------------------------------------------------------------------------------------------------------------------------
	// Helpers
	
	/**
	 * 
	 */
	private function download_files( $font, $subset, $variants )
	{
		// define url
		$query = http_build_query(array(
			'download' 	=> 'zip',
			'subsets' 	=> implode(',', $subset),				
			'variants' 	=> implode(',', $variants),
			'formats' 	=> 'woff,woff2'
		));

		// get wordpress upload path
		$wp_upload = wp_get_upload_dir();

		// define temp file
		$temp_file = $wp_upload['basedir'] . '/' . $font . '.zip';

		// open new file stream
		$temp_file_stream = fopen($temp_file, 'w+');

		// make api request
		$response = wp_remote_get("https://google-webfonts-helper.herokuapp.com/api/fonts/" . $font . "?" . $query);

		// write body to file stream
		fwrite($temp_file_stream, wp_remote_retrieve_body($response));

		// close file stream
		fclose($temp_file_stream);

		//
		return $temp_file;
	}

	/**
	 * 
	 */
	private function unzip_file( $zip_file )
	{
		// init wordpress filesystem
		require_once(ABSPATH .'/wp-admin/includes/file.php');
		global $wp_filesystem;
		WP_Filesystem();

		// get wordpress upload path
		$wp_upload = wp_get_upload_dir();

		// 
		$unzipped_files = unzip_file( $zip_file, $wp_upload['basedir'] . '/' . $this->plugin_name . '/fonts');

		// delete uploaded file
		unlink($zip_file);

		//
		return $unzipped_files;
	}

    /**
     * 
     */
    private function define_output_directions($slug) 
    {
        //
        $result = array();

		$wp_upload = wp_get_upload_dir();
		$result['path'] = $wp_upload['basedir'] . '/' . $slug;
        $result['url']  = $wp_upload['baseurl'] . '/' . $slug;
        
        //
        return $result;
    }

	/**
	 * 
	 */
	private function generate_style()
	{
		//
		$files = $this->get_fonts_files();

		//
		$result = "";

		//
		foreach($files as $font_id => $font_weights)
		{
			//
			$font_family = ucwords(str_replace('-', ' ', $font_id));

			//
			foreach($font_weights as $font_weight => $font_styles)
			{
				//
				foreach($font_styles as $font_style => $font_files)
				{
					// open
					$result .= "/* " . $font_id . " - " . $font_weight . " - " . $font_style . " */\n";
					$result .= "@font-face {" . "\n";

					//
					$result .= "\tfont-family: '" . $font_family . "';" . "\n";
					$result .= "\tfont-style: " . $font_style . ";" . "\n";
					$result .= "\tfont-weight: " . $font_weight . ";" . "\n";
					$result .= "\tsrc: local('')," . "\n";

					//
					$total = count($font_files);
					
					//
					for ($i = 0; $i < $total; $i++) 
					{
						// extract extension from file name
						$extension = explode('.', $font_files[$i]);
						$extension = end( $extension );
						
						//
						$result .= "\t\turl('fonts/" . $font_files[$i] . "') format('" . $extension . "')" . ($i == ($total - 1) ? ';' : ',') . "\n";
					}

					// close
					$result .= "}" . "\n";
					$result .= "\n";
				}
			}	
		}

		//
		return $result;
	}

	/**
	 *
	 */
	private function get_fonts_files()
	{
		// define result
		$result = array();

		// get all files in wp-content\uploads\hassle-free-fonts\fonts
		foreach(array_slice(scandir($this->output['path'] . '/fonts'), 2) as $fonts_file) 
		{
			// get values
			$font = $this->extract_font_value($fonts_file);

			// if key not is defined create a new entry
			if(!array_key_exists($font->id, $result)) {
				$result[$font->id] = array();
			}

			//
			if(!array_key_exists($font->weight, $result[$font->id])) {
				$result[$font->id][$font->weight] = array();
			}

			//
			if(!array_key_exists($font->style, $result[$font->id][$font->weight])) {
				$result[$font->id][$font->weight][$font->style] = array();
			}

			//
			array_push($result[$font->id][$font->weight][$font->style], $fonts_file);
		}

		//
		return $result;
	}

	/**
	 * 
	 */
	private function extract_font_value($file_name)
	{
		//
		$result = new stdClass();

		// split filename to extract id, weight and style
		$font_pieces = explode("-v", $file_name);

		// first in array is font_id
		$result->id = $font_pieces[0];

		//
		$result->family = ucwords(str_replace('-', ' ', $result->id));

		// extract from last piece into weight and style
		$font_weight_style = explode("-", end($font_pieces));
		$font_weight_style = end($font_weight_style);
		$font_weight_style = explode(".", $font_weight_style);
		$font_weight_style = $font_weight_style[0];

		//
		if ( strpos($font_weight_style, 'italic') !== false ) 
		{
			//
			$result->weight = str_replace("italic","", $font_weight_style);
			$result->style = "italic";
		}

		//
		elseif ( strpos($font_weight_style, 'regular') !== false ) 
		{
			//
			$result->weight = "400";
			$result->style = "normal";
		}

		//
		else 
		{
			//								
			$result->weight = $font_weight_style;
			$result->style = "normal";
		}

		//
		if($result->weight == "") {
			$result->weight = "400";
		}

		//
		return $result;
	}

    //--------------------------------------------------------------------------------------------------------------------------------
}